#ifndef EEPROMmgr_cpp
#define EEPROMmgr_cpp

#include "EEPROMmgr.h"
#include "../INJducatimgr/INJducatimgr.h" // To read and write injection percentage increase calibration maps

// EEPROM memory addresses table:
#define INJ_INCR_RPM_MAPS_START 0 // 0
#define INJ_INCR_TIM_MAPS_START 4*INJ_INCR_RPM_MAPS_SIZE // 32
#define SD_FILE_NUM_ADDR 64 // address of file number (4 bytes used)
#define EEPROM_CONFIG_WORD_ADDR 68 // address of Configuration Word
#define BATTERY_CHECK_BYPASS_BIT 0
#define ENG_DATA_LOG_BYPASS_BIT 1
#define GPS_DATA_LOG_BYPASS_BIT 2
#define IMU_DATA_LOG_BYPASS_BIT 3
#define LAM_DATA_LOG_BYPASS_BIT 4
#define ALL_DATA_LOG_BYPASS_BIT 5

#define INJ_INCREMENT_RPM_STD_LO (uint8_t)0
#define INJ_INCREMENT_RPM_STD_HI (uint8_t)0
#define INJ_INCREMENT_TIM_STD_LO (uint8_t)0xE8
#define INJ_INCREMENT_TIM_STD_HI (uint8_t)0x03

uint8_t EEPROM_config_word = 0; // Configuration word saved into EEPROM ("0x00" is the temporary value, in case of fault)

uint8_t bat_check_inhibit(){ return ((EEPROM_config_word & (1 << BATTERY_CHECK_BYPASS_BIT)) >> BATTERY_CHECK_BYPASS_BIT); }
uint8_t eng_log_inhibit(){ return ((EEPROM_config_word & (1 << ENG_DATA_LOG_BYPASS_BIT)) >> ENG_DATA_LOG_BYPASS_BIT); }
uint8_t gps_log_inhibit(){ return ((EEPROM_config_word & (1 << GPS_DATA_LOG_BYPASS_BIT)) >> GPS_DATA_LOG_BYPASS_BIT); }
uint8_t imu_log_inhibit(){ return ((EEPROM_config_word & (1 << IMU_DATA_LOG_BYPASS_BIT)) >> IMU_DATA_LOG_BYPASS_BIT); }
uint8_t lam_log_inhibit(){ return ((EEPROM_config_word & (1 << LAM_DATA_LOG_BYPASS_BIT)) >> LAM_DATA_LOG_BYPASS_BIT); }
uint8_t all_log_inhibit(){ return ((EEPROM_config_word & (1 << ALL_DATA_LOG_BYPASS_BIT)) >> ALL_DATA_LOG_BYPASS_BIT); }

// Loads the configuration word from the EEPROM
uint8_t EEPROM_load_config_word(){
	uint8_t config_word_tmp = EEPROM.read(EEPROM_CONFIG_WORD_ADDR); // original value
	uint8_t config_word_tmp_red = EEPROM.read(EEPROM_CONFIG_WORD_ADDR+1); // redundant value
	if (config_word_tmp_red == ((uint8_t)255 - config_word_tmp)){ // checksum check OK
		EEPROM_config_word = config_word_tmp;
		return 1; // OK
	}
	return 0; // NG
}


// Writes injection standard values into EEPROM
uint8_t EEPROM_write_standard_values(uint8_t data_number){
  uint8_t i;
  if ((data_number ==0) || (data_number ==INJ_MAPS_TOTAL_NUM)){ // RPM map
	  for (i=0; i<INJ_INCR_RPM_MAPS_SIZE; i++){
		  EEPROM.write(INJ_INCR_RPM_MAPS_START+2*i, INJ_INCREMENT_RPM_STD_LO); //main data
		  EEPROM.write(INJ_INCR_RPM_MAPS_START+2*i+1, INJ_INCREMENT_RPM_STD_HI); //main data
		  EEPROM.write(INJ_INCR_RPM_MAPS_START+2*(INJ_INCR_RPM_MAPS_SIZE+i), (uint8_t)255 - INJ_INCREMENT_RPM_STD_LO); //for redundancy, this is a copy of the main data
		  EEPROM.write(INJ_INCR_RPM_MAPS_START+2*(INJ_INCR_RPM_MAPS_SIZE+i)+1, (uint8_t)255 - INJ_INCREMENT_RPM_STD_HI); //for redundancy, this is a copy of the main data
		  incrementi_rpm[i]=((uint16_t)INJ_INCREMENT_RPM_STD_HI << 8) | (uint16_t)INJ_INCREMENT_RPM_STD_LO; // take the standard value
	  }
  }
  if ((data_number ==1) || (data_number ==INJ_MAPS_TOTAL_NUM)){ // THR map
	for (i=0; i<INJ_INCR_TIM_MAPS_SIZE; i++){
		  EEPROM.write(INJ_INCR_TIM_MAPS_START+2*i, INJ_INCREMENT_TIM_STD_LO); //main data
		  EEPROM.write(INJ_INCR_TIM_MAPS_START+2*i+1, INJ_INCREMENT_TIM_STD_HI); //main data
		  EEPROM.write(INJ_INCR_TIM_MAPS_START+2*(INJ_INCR_TIM_MAPS_SIZE+i), (uint8_t)255 - INJ_INCREMENT_TIM_STD_LO); //for redundancy, this is a copy of the main data
		  EEPROM.write(INJ_INCR_TIM_MAPS_START+2*(INJ_INCR_TIM_MAPS_SIZE+i)+1, (uint8_t)255 - INJ_INCREMENT_TIM_STD_HI); //for redundancy, this is a copy of the main data
		  incrementi_tim[i]=((uint16_t)INJ_INCREMENT_TIM_STD_HI << 8) | (uint16_t)INJ_INCREMENT_TIM_STD_LO; // take the standard value
	  }
  }
  return 0;
}


// Loads calibration values for injection. If an error is found, standard values are written
uint8_t EEPROM_load_calib(){
	uint8_t error_flag = 0;
	for (uint8_t i=0; i<INJ_INCR_TIM_MAPS_SIZE; i++){ // Imports "incrementi_tim"
		uint8_t tmp_L = EEPROM.read(INJ_INCR_TIM_MAPS_START+2*i);
		uint8_t tmp_H = EEPROM.read(INJ_INCR_TIM_MAPS_START+2*i+1);
		uint16_t tmp_val = ((uint16_t)tmp_H << 8) | (uint16_t)tmp_L;
		tmp_L = EEPROM.read(INJ_INCR_TIM_MAPS_START+2*INJ_INCR_TIM_MAPS_SIZE+2*i);
		tmp_H = EEPROM.read(INJ_INCR_TIM_MAPS_START+2*INJ_INCR_TIM_MAPS_SIZE+2*i+1);
		uint16_t tmp_val_red = ((uint16_t)tmp_H << 8) | (uint16_t)tmp_L;
		if (tmp_val != ((uint16_t)0xFFFF - tmp_val_red)){ // redundancy check NG
			error_flag=1;
		}
		else{ // OK
			incrementi_tim[i]=tmp_val;
		}
	}
	for (uint8_t i=0; i<INJ_INCR_RPM_MAPS_SIZE; i++){ // Imports "incrementi_tim"
		uint8_t tmp_L = EEPROM.read(INJ_INCR_RPM_MAPS_START+2*i);
		uint8_t tmp_H = EEPROM.read(INJ_INCR_RPM_MAPS_START+2*i+1);
		uint16_t tmp_val = ((uint16_t)tmp_H << 8) | (uint16_t)tmp_L;
		tmp_L = EEPROM.read(INJ_INCR_RPM_MAPS_START+2*INJ_INCR_RPM_MAPS_SIZE+2*i);
		tmp_H = EEPROM.read(INJ_INCR_RPM_MAPS_START+2*INJ_INCR_RPM_MAPS_SIZE+2*i+1);
		uint16_t tmp_val_red = ((uint16_t)tmp_H << 8) | (uint16_t)tmp_L;
		if (tmp_val != ((uint16_t)0xFFFF - tmp_val_red)){ // redundancy check NG
			error_flag=1;
		}
		else{ // OK
			incrementi_rpm[i]=tmp_val;
		}
	}
	if (error_flag){
		EEPROM_write_standard_values(INJ_MAPS_TOTAL_NUM); // Writes the standard values
		return 0; // Not OK
	}
	return 1; // OK
}


// Reads data from EEPROM, and initializes them in RAM memory
uint8_t EEPROM_initialize(){
	uint8_t error_flag = 0;
	error_flag = EEPROM_load_config_word(); // Loads config word
	error_flag = EEPROM_load_calib(); // Loads the calibration maps
	return error_flag;
}


// Writes a map from RAM into EEPROM memory
void EEPROM_write_RAM_map_to_EEPROM(uint8_t map_number_req){
	uint16_t value_tmp = 0;
	uint8_t value_tmp_L = 0;
	uint8_t value_tmp_H = 0;
	if ((map_number_req ==0) || (map_number_req == INJ_MAPS_TOTAL_NUM)){ // rpm
		for (uint8_t i=0; i<INJ_INCR_RPM_MAPS_SIZE;i++){
			value_tmp = incrementi_rpm[i];
			value_tmp_L = (uint8_t)(value_tmp & (uint16_t)0x00FF); // low
			value_tmp_H = (uint8_t)(value_tmp >> 8); // high
			EEPROM.write(INJ_INCR_RPM_MAPS_START+2*i, value_tmp_L); //main data
			EEPROM.write(INJ_INCR_RPM_MAPS_START+2*i+1, value_tmp_H); //main data
			EEPROM.write(INJ_INCR_RPM_MAPS_START+2*(INJ_INCR_RPM_MAPS_SIZE+i), (uint8_t)255 - value_tmp_L); //for redundancy, this is a copy of the main data
			EEPROM.write(INJ_INCR_RPM_MAPS_START+2*(INJ_INCR_RPM_MAPS_SIZE+i)+1, (uint8_t)255 - value_tmp_H); //for redundancy, this is a copy of the main data
		}
	}
	if ((map_number_req ==1) || (map_number_req == INJ_MAPS_TOTAL_NUM)){ // injection time
		for (uint8_t i=0; i<INJ_INCR_TIM_MAPS_SIZE;i++){
			value_tmp = incrementi_tim[i];
			value_tmp_L = (uint8_t)(value_tmp & (uint16_t)0x00FF); // low
			value_tmp_H = (uint8_t)(value_tmp >> 8); // high
			EEPROM.write(INJ_INCR_TIM_MAPS_START+2*i, value_tmp_L); //main data
			EEPROM.write(INJ_INCR_TIM_MAPS_START+2*i+1, value_tmp_H); //main data
			EEPROM.write(INJ_INCR_TIM_MAPS_START+2*(INJ_INCR_TIM_MAPS_SIZE+i), (uint8_t)255 - value_tmp_L); //for redundancy, this is a copy of the main data
			EEPROM.write(INJ_INCR_TIM_MAPS_START+2*(INJ_INCR_TIM_MAPS_SIZE+i)+1, (uint8_t)255 - value_tmp_H); //for redundancy, this is a copy of the main data
		}
	}
}


// Reads file number into EEPROM, increases it (+1), writes the increased number into EEPROM, and returns that value. In case of error, "0" is used.
uint16_t EEPROM_SD_file_num_rw(){
	
	// EEPROM values read
	uint8_t valoreL = EEPROM.read(SD_FILE_NUM_ADDR); // original value
	uint8_t valoreH = EEPROM.read(SD_FILE_NUM_ADDR+1); // original value
	uint8_t valore_redL = EEPROM.read(SD_FILE_NUM_ADDR+2); // redundancy
	uint8_t valore_redH = EEPROM.read(SD_FILE_NUM_ADDR+3); // redundancy
	
	// Deciding values
	if ((valore_redH == ((uint8_t)255 - valoreH)) && (valore_redL == ((uint8_t)255 - valoreL))){ // checksum check OK
		uint16_t valore = ((uint16_t)valoreL + ((uint16_t)valoreH << 8));
		valore++; // increase value
		valoreL = (uint8_t)(valore & (uint16_t)0x00FF); // low
		valoreH = (uint8_t)(valore >> 8); // high
	}else{ // initialization (standard values)
		valoreL = (uint8_t)0;
		valoreH = (uint8_t)0;
	}
	valore_redL=(uint8_t)255 - valoreL;
	valore_redH=(uint8_t)255 - valoreH;
	
	// Writing new values into EEPROM
	EEPROM.write(SD_FILE_NUM_ADDR, valoreL);
	EEPROM.write(SD_FILE_NUM_ADDR+1, valoreH);
	EEPROM.write(SD_FILE_NUM_ADDR+2, valore_redL);
	EEPROM.write(SD_FILE_NUM_ADDR+3, valore_redH);
	
	return ((uint16_t)valoreL + ((uint16_t)valoreH << 8));
}


#endif