#ifndef INJducatimgr_cpp
#define INJducatiINJmgr_cpp

#include <Arduino.h>

#include "INJducatimgr.h"
#include "tempodouble/tempodouble.h"
#include "../COMMmgr/COMMmgr.h"
#include "../ADCmgr/ADCmgr.h"

#define N_CYLS_ACTIVE 2 // number of cylinders (i.e. injectors)
#define IN_INJ1_PIN 6 // input pin injector from ECU
#define OUT_INJ1_PIN 8 // output pin injector to physical injector
#define IN_INJ2_PIN 7 // input pin injector from ECU
#define OUT_INJ2_PIN 9 // output pin injector to physical injector

#define INJ_PIN_FILT_MEASURES (uint8_t)8 // filtering number
#define INJ_PIN_FILT_LOW (uint8_t)2
#define INJ_PIN_FILT_HIGH (uint8_t)6
#define INJ_TIME_TICKS_MIN (uint16_t)50 // minimum injection 200us (1 Timer0 tick = 4us)
#define INJ_TIME_TICKS_MAX (uint16_t)2000 // maximmum injection 8ms = 8000us (1 Timer0 tick = 4us)
#define INJ_OPENING_TIME_TICKS (uint16_t)0 // time ticks (4us) required for the injector to open [example: 6 == 24us]
#define MAX_PERCENTAGE_INJ (uint16_t)32768 // maximum increment percentage (for safety) [32768 corresponds to 50% increment, because 2e16 = 100%]
#define MIN_EXTENSION_TIME_TICKS (uint16_t)100 // minimum time of Timer1 extension time ticks (0.5us), for injection time extension [100 = 50us]
#define MAX_EXTENSION_TIME_TICKS (uint16_t)6000 // minimum time of Timer1 extension time ticks (0.5us), for injection time extension [6000 = 3000us]
#define INJ_CALCULATION_TIME_TICKS (uint16_t)54 // time ticks (0.5us) required for calculation of extension time [example: 60 == 30us]

// GLOBAL VARIABLES
volatile uint8_t engine_running_flag = 1; // Becomes ON when the engine is cranked
volatile uint16_t inj_start_tick_last[N_CYLS_ACTIVE]; // time at which injection was started last time
volatile uint16_t injection_counter[N_CYLS_ACTIVE]; // counts the combustion cycles
volatile uint16_t delta_time_tick[N_CYLS_ACTIVE]; // time between 2 injections (2 rotations)
volatile uint16_t delta_inj_tick[N_CYLS_ACTIVE]; // injection time input
volatile uint16_t extension_time_ticks[N_CYLS_ACTIVE]; // extension time Timer1 ticks (0.5us)
volatile uint16_t perc_inc[N_CYLS_ACTIVE]; // increment percentage
volatile uint8_t INJ_exec_time = 0; // interrupt execution time
volatile uint16_t incrementi_rpm[INJ_INCR_RPM_MAPS_SIZE]; // increments, based on rpm
const uint16_t incrementi_rpm_brkpts[] = {2600, 3112, 4136, 5160, 7208, 11304, 15400, 23592}; // breakpoints, size should be INJ_INCR_RPM_MAPS_SIZE
const uint8_t incrementi_rpm_shifts[] = {9, 10, 10, 11, 12, 12, 13}; // breakpoints, size should be (INJ_INCR_RPM_MAPS_SIZE-1)
volatile uint16_t incrementi_tim[INJ_INCR_TIM_MAPS_SIZE]; // increments, based on injection time
const uint16_t incrementi_tim_brkpts[] = {256, 512, 768, 1024, 1280, 1536, 1792, 2048}; // breakpoints, size should be INJ_INCR_TIM_MAPS_SIZE
const uint8_t incrementi_tim_shifts[] = {8, 8, 8, 8, 8, 8, 8}; // breakpoints, size should be (INJ_INCR_TIM_MAPS_SIZE-1)

// Buffer variables, to store data before sending on Serial or storing on SD
volatile uint16_t injection_counter_buffer=0; // counts the combustion cycles
volatile uint16_t delta_time_tick_buffer=0; // time between 2 injections buffer
volatile uint16_t inj_start_tick_last_buffer=0; // time of start of injection
volatile uint16_t delta_inj_tick_buffer=0; // time of injection input buffer
volatile uint16_t extension_time_ticks_buffer=0; // extension time ticks
volatile uint8_t buffer_busy=0; // activated when the data are being read by other functions

uint8_t packet_cnt=0;

// Buffers to memorize combustion data
struct combustion_buffer_struct{
  uint16_t injection_counter;
  uint32_t inj_start_ticks;
  uint32_t inj_end_in_ticks;
  uint32_t inj_end_out_ticks;
  //uint16_t throttle;
  //uint16_t lambda;
  uint16_t temperature;
  //uint8_t INJ_exec_time_1;
  //uint8_t INJ_exec_time_2;
};
volatile combustion_buffer_struct combustion_buffers_1[INJ_COMBUSTION_BUFFERS_NUM]; // combustion buffer for cylinder 1
volatile combustion_buffer_struct combustion_buffers_2[INJ_COMBUSTION_BUFFERS_NUM]; // combustion buffer for cylinder 2
volatile uint8_t buffer_last_fill = 0;
volatile uint8_t buffer_last_write = 0;
volatile uint8_t buffer_now_fill = 0;
volatile uint8_t buffer_now_write = 0;
volatile bool buffer_write_possible = false;

enum INJ_signal_trigger_mode_enum{
	WAIT_FOR_ON = 0,
	WAIT_FOR_OFF
};
volatile INJ_signal_trigger_mode_enum INJ_signal_trigger_mode[N_CYLS_ACTIVE];


// FUNCTIONS

void interrupt_inj_1_off(){
	digitalWrite(OUT_INJ1_PIN, LOW); // shuts down injector
	if (buffer_write_possible){
		combustion_buffers_1[buffer_now_fill].inj_end_out_ticks = Timer0_tick_counts(); // update buffer
		buffer_last_fill = buffer_now_fill; // filling completed
		buffer_write_possible = false; // not possible to write
	}
	if (!buffer_busy){ // we are into an interrupt, and we must make sure that the normal execution is not accessing this data
		injection_counter_buffer = injection_counter[0];
		delta_time_tick_buffer = delta_time_tick[0];
		delta_inj_tick_buffer = delta_inj_tick[0];
		extension_time_ticks_buffer = extension_time_ticks[0];
	}
	INJ_signal_trigger_mode[0] = WAIT_FOR_ON; // Waiting for next ON event
}


void interrupt_inj_2_off(){
	digitalWrite(OUT_INJ2_PIN, LOW); // shuts down injector
	INJ_signal_trigger_mode[1] = WAIT_FOR_ON; // Waiting for next ON event
}


void INJducatimgr_class::begin(){
	
	Timer1.initialize(); // Initializes Timer1 for injection time management
	Timer1.attachInterrupt(interrupt_inj_1_off, 0);
	Timer1.attachInterrupt(interrupt_inj_2_off, 1);
	Timer1.start();
	
	PCICR |=  1 << PCIE2; // PCIE2 (Port D)
	pinMode(IN_INJ1_PIN, INPUT_PULLUP);
	pinMode(OUT_INJ1_PIN, OUTPUT);
	digitalWrite(OUT_INJ1_PIN, LOW);
	PCMSK2 |= 1 << PCINT22; // PCINT22 (Port D Pin 6)
	#if N_CYLS_ACTIVE == 2
		pinMode(IN_INJ2_PIN, INPUT_PULLUP);
		pinMode(OUT_INJ2_PIN, OUTPUT);
		digitalWrite(OUT_INJ2_PIN, LOW);
		PCMSK2 |= 1 << PCINT23; // PCINT23 (Port D Pin 7)
		
	#endif
	for (uint8_t i=0; i<N_CYLS_ACTIVE; i++){
		INJ_signal_trigger_mode[i] = WAIT_FOR_ON;
		inj_start_tick_last[i]=0; // time at which injection was started last time
		injection_counter[i]=0; // counts the combustion cycles
		delta_time_tick[i]=0; // time between 2 injections (2 rotations)
		delta_inj_tick[i]=0; // injection time input
		perc_inc[i]=0; // increment percentage
		extension_time_ticks[i]=0; // extension time Timer1 ticks (0.5us)
	}
	PCIFR |= 1 << PCIF2; // PCIF2 (Port D)
	
}


// Returns the number of tick counts of Timer0 (1 tick = 4 us, since f=16MHz and prescaler = 64). 1 Overflow happens each 256 * 4 us = 1024 us
// The returned data size is 2 bytes (16 bit), allowing to count up to 262144 us
extern volatile unsigned long timer0_overflow_count; // needed to read the microseconds counter
uint32_t Timer0_tick_counts() {
	uint32_t m;
	uint8_t t;
	uint8_t oldSREG = SREG;
	cli();
	m = (uint32_t)timer0_overflow_count;
	t = TCNT0;
	if ((TIFR0 & _BV(TOV0)) && (t < 255)) m++;
	SREG = oldSREG;
	return ((m << 8) + (uint32_t)t);
}


// Interpolates the map of fuel injection increment time, as a function of input injection time
uint16_t INJducatimgr_class::interpolate_tim_map(uint16_t inj_ticks_in){
	
	// Search for proper index
	if (inj_ticks_in <= incrementi_tim_brkpts[0]) return incrementi_tim[0]; // return array lowest value (the x value was lower than all the elements)
	uint8_t index; // index found	
	for (index=1; index<INJ_INCR_TIM_MAPS_SIZE; index++){ // scan all breakpoints
		if (inj_ticks_in < incrementi_tim_brkpts[index]){ // x lower is higher than next element
			index--; // previous index
			break; // exit cycle
		}
	}
	if (index >= (INJ_INCR_TIM_MAPS_SIZE - 1)) return incrementi_tim[INJ_INCR_TIM_MAPS_SIZE-1]; // return array highest value (the x value was higher than all the elements)
	
	// Performs subtraction and multiplication of the remaining part
	uint16_t rem_x = inj_ticks_in - incrementi_tim_brkpts[index]; // remaining part of x value
	uint16_t y0 = incrementi_tim[index]; // Low Y element
	uint16_t y1 = incrementi_tim[index+1]; // High Y element
	uint16_t dy; // "dy" for slope calculation
	bool positive; // "dy" is positive flag
	if (y1 >= y0){
		dy = y1 - y0; // y increment positive
		positive=true;
	}else{
		dy = y0 - y1; // y increment negative
		positive=false;
	}
	uint32_t mult_tmp = ((uint32_t)rem_x * (uint32_t)dy); // multiplication between 16 bits number and 16 bit number
	
	// Performs division
	uint8_t div_shifts = incrementi_tim_shifts[index]; // number of shifts on the right (division by "dx")
	for (uint8_t i=0; i<div_shifts; i++){ // shift right "div_shift" times
		mult_tmp = mult_tmp >> 1; 
	}
		
	// Performs addition
	if (positive == true){
		return (y0 + (uint16_t)mult_tmp);
	}else{
		return (y0 + (uint16_t)mult_tmp); // 1e8 = 50% ==> 1e16 = 100%
	}
}


// Interpolates the rpm map and calculates, as output, the injection time increment (%). 2 bytes. Resolution: 2e16 = 50%
uint16_t INJducatimgr_class::interpolate_rpm_map(uint16_t inj_2r_ticks_in){

/*	
	// Search for proper index
	if (inj_2r_ticks_in <= incrementi_rpm_brkpts[0]) return incrementi_rpm[0]; // return array lowest value (the x value was lower than all the elements)
	uint8_t index; // index found
	for (index=1; index<INJ_INCR_RPM_MAPS_SIZE; index++){ // scan all breakpoints
		if (delta_time_tick < incrementi_rpm_brkpts[index]){ // x lower is higher than next element
			index--; // previous index
			break; // exit cycle
		}
	}
	if (index >= (INJ_INCR_RPM_MAPS_SIZE - 1)) return ((uint16_t)incrementi_rpm[INJ_INCR_RPM_MAPS_SIZE-1] << 7); // return array highest value (the x value was higher than all the elements)
	
	// Performs subtraction and multiplication of the remaining part
	uint16_t rem_x = delta_time_tick - incrementi_rpm_brkpts[index]; // remaining part of x value
	uint8_t y0 = incrementi_rpm[index]; // Low Y element
	uint8_t y1 = incrementi_rpm[index+1]; // High Y element
	uint8_t dy; // "dy" for slope calculation
	bool positive; // "dy" is positive flag
	if (y1 >= y0){
		dy = y1 - y0; // y increment positive
		positive=true;
	}else{
		dy = y0 - y1; // y increment negative
		positive=false;
	}
	uint32_t mult_tmp = ((uint32_t)rem_x * (uint32_t)dy); // multiplication between 16 bits number and 8 bit number
	
	// Performs division and resolution shifting
	uint8_t div_shifts = incrementi_rpm_shifts[index]; // number of shifts on the right (division by "dx")
	if (div_shifts >= 7){ // Division is bigger than multiplication
		for (uint8_t i=0; i<(div_shifts - 7); i++){ // shift right "div_shift" times. "7" multiplication is to present the value in "10e16 = 100%" format
			mult_tmp = mult_tmp >> 1; 
		}
	}else{
		for (uint8_t i=0; i<(7 - div_shifts); i++){ // shift right "div_shift" times. "7" multiplication is to present the value in "10e16 = 100%" format
			mult_tmp = mult_tmp << 1; 
		}
	}
	
	// Performs addition
	if (positive == true){
		return ((uint16_t)y0 << 7) + (uint16_t)mult_tmp; // 1e8 = 50% ==> 1e16 = 100%
	}else{
		return ((uint16_t)y0 << 7) - (uint16_t)mult_tmp; // 1e8 = 50% ==> 1e16 = 100%
	}
*/
	return 0; // temporary
}


// Creates the packet of Engine Data, for SD card logging
uint8_t INJducatimgr_class::create_engine_data_packet(uint8_t* SD_writing_buffer, uint8_t max_length){
	
	// Preparation of Engine Data array (this is needed also for Serial communication - service protocol)
	uint8_t i=0; // packet bytes counter
	SD_writing_buffer[i++] = 'd';
	SD_writing_buffer[i++] = packet_cnt; // packet counter
	packet_cnt++; // increase packet counter
	unsigned long time_stamp_temp = millis(); // read time stamp
	SD_writing_buffer[i++] = (uint8_t)(time_stamp_temp & 0xff); // LSB
    SD_writing_buffer[i++] = (uint8_t)((time_stamp_temp >> 8) & 0xff);
    SD_writing_buffer[i++] = (uint8_t)((time_stamp_temp >> 16) & 0xff);
    SD_writing_buffer[i++] = (uint8_t)((time_stamp_temp >> 24) & 0xff); // MSB
	buffer_busy = 1; // Locks the buffer access (semaphore)
	SD_writing_buffer[i++] = (uint8_t)(injection_counter_buffer & 0xff); // LSB
	SD_writing_buffer[i++] = (uint8_t)((injection_counter_buffer >> 8) & 0xff); // MSB
    SD_writing_buffer[i++] = (uint8_t)(inj_start_tick_last_buffer & 0xff); // LSB
	SD_writing_buffer[i++] = (uint8_t)((inj_start_tick_last_buffer >> 8) & 0xff); // MSB
	SD_writing_buffer[i++] = (uint8_t)(delta_time_tick_buffer & 0xff); // LSB
	SD_writing_buffer[i++] = (uint8_t)((delta_time_tick_buffer >> 8) & 0xff); // MSB
    SD_writing_buffer[i++] = (uint8_t)(delta_inj_tick_buffer & 0xff); // LSB
    SD_writing_buffer[i++] = (uint8_t)((delta_inj_tick_buffer >> 8) & 0xff); // MSB
    //SD_writing_buffer[i++] = (uint8_t)(throttle_buffer & 0xff); // LSB
    //SD_writing_buffer[i++] = (uint8_t)((throttle_buffer >> 8) & 0xff); // MSB
	//SD_writing_buffer[i++] = (uint8_t)(lambda_buffer & 0xff); // LSB
    //SD_writing_buffer[i++] = (uint8_t)((lambda_buffer >> 8) & 0xff); // MSB
	SD_writing_buffer[i++] = 0x00;
	SD_writing_buffer[i++] = 0x00;
	SD_writing_buffer[i++] = 0x00;
	SD_writing_buffer[i++] = 0x00;
	SD_writing_buffer[i++] = (uint8_t)(extension_time_ticks_buffer & 0xff); // LSB
	SD_writing_buffer[i++] = (uint8_t)((extension_time_ticks_buffer >> 8) & 0xff); // MSB
	buffer_busy = 0; // Opens the buffer again
	SD_writing_buffer[i++] = (uint8_t)INJ_exec_time; // LSB
	SD_writing_buffer[i++] = (uint8_t)INJ_exec_time; // LSB
	SD_writing_buffer[i++] = ADCmgr_binary_inputs_status_read(); // digital inputs status
    uint16_t CK_SUM = COMM_calculate_checksum(SD_writing_buffer, 0, i);
    SD_writing_buffer[i++] = (uint8_t)(CK_SUM >> 8);
    SD_writing_buffer[i++] = (uint8_t)(CK_SUM & 0xFF);
	
}


// Creates the packet of Combustion Data, for SD card logging and Comm
void INJducatimgr_class::create_combustion_data_packet(){
	
	volatile combustion_buffer_struct* combustion_buffers = combustion_buffers_1; // set pointer to Cylinder 1
	uint8_t i=0; // packet bytes counter
	SD_combustion_buffer[i++] = 'f'; // fire
	SD_combustion_buffer[i++] = packet_cnt; // packet count
	packet_cnt++; // increase packet counter
	unsigned long time_stamp_temp = millis(); // read time stamp
	SD_combustion_buffer[i++] = (uint8_t)(time_stamp_temp & 0xff); // LSB
    SD_combustion_buffer[i++] = (uint8_t)((time_stamp_temp >> 8) & 0xff);
    SD_combustion_buffer[i++] = (uint8_t)((time_stamp_temp >> 16) & 0xff);
    SD_combustion_buffer[i++] = (uint8_t)((time_stamp_temp >> 24) & 0xff); // MSB
	//SD_combustion_buffer[i++] = ADCmgr_binary_inputs_status_read(); // digital inputs status
	buffer_now_write = buffer_last_write+1;
	if (buffer_now_write==INJ_COMBUSTION_BUFFERS_NUM) buffer_now_write=0; // rollout
	SD_combustion_buffer[i++] = (uint8_t)(combustion_buffers[buffer_now_write].injection_counter & 0xff); // LSB
    SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].injection_counter >> 8) & 0xff);
	SD_combustion_buffer[i++] = (uint8_t)(combustion_buffers[buffer_now_write].inj_start_ticks & 0xff); // LSB
    SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].inj_start_ticks >> 8) & 0xff);
    SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].inj_start_ticks >> 16) & 0xff);
    SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].inj_start_ticks >> 24) & 0xff); // MSB
	SD_combustion_buffer[i++] = (uint8_t)(combustion_buffers[buffer_now_write].inj_end_in_ticks & 0xff); // LSB
    SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].inj_end_in_ticks >> 8) & 0xff);
    SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].inj_end_in_ticks >> 16) & 0xff);
    SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].inj_end_in_ticks >> 24) & 0xff); // MSB
	SD_combustion_buffer[i++] = (uint8_t)(combustion_buffers[buffer_now_write].inj_end_out_ticks & 0xff); // LSB
    SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].inj_end_out_ticks >> 8) & 0xff);
    SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].inj_end_out_ticks >> 16) & 0xff);
    SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].inj_end_out_ticks >> 24) & 0xff); // MSB
	//SD_combustion_buffer[i++] = (uint8_t)(combustion_buffers[buffer_now_write].throttle & 0xff); // LSB
	//SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].throttle >> 8) & 0xff); // MSB
	//SD_combustion_buffer[i++] = (uint8_t)(combustion_buffers[buffer_now_write].lambda & 0xff); // LSB
	//SD_combustion_buffer[i++] = (uint8_t)((combustion_buffers[buffer_now_write].lambda >> 8) & 0xff); // MSB
	uint16_t temp_tmp = ADCmgr_temperature_signal_read();
	SD_combustion_buffer[i++] = (uint8_t)(temp_tmp & 0xff); // LSB
	SD_combustion_buffer[i++] = (uint8_t)((temp_tmp >> 8) & 0xff); // MSB
	//SD_combustion_buffer[i++] = (uint8_t)INJ_exec_time; // LSB
	//SD_combustion_buffer[i++] = (uint8_t)combustion_buffers[buffer_now_write].INJ_exec_time_2; // LSB
    uint16_t CK_SUM = COMM_calculate_checksum(SD_combustion_buffer, 0, i);
    SD_combustion_buffer[i++] = (uint8_t)(CK_SUM >> 8);
    SD_combustion_buffer[i++] = (uint8_t)(CK_SUM & 0xFF);
	
}


// Interrupt pins status changes (Original ECU injector command pin). This routine is called at any change on PortD pin.
ISR(PCINT2_vect){

	// For execution time measurement
	uint32_t INJ_exec_time_start = Timer0_tick_counts(); // 1 tick = 4us
	
	// Evaluate the status of the INJ input pins
	uint8_t INJ_status_filtered[2]; // status after filtering
	INJ_status_filtered[0]=2; // Injector pin status (Low = ON, High = OFF), 2 -> not defined
	INJ_status_filtered[1]=2; // Injector pin status (Low = ON, High = OFF), 2 -> not defined
	uint8_t pin_status_cnt[2];
	pin_status_cnt[0] = 0;
	pin_status_cnt[1] = 0;
	for (uint8_t i=0; i<INJ_PIN_FILT_MEASURES; i++){ // Performs many measurements and accumulates the status (0 or 1) in a counter
		pin_status_cnt[0] += ((PIND & _BV(IN_INJ1_PIN)) >> IN_INJ1_PIN); // injector 1
		if (N_CYLS_ACTIVE==2) pin_status_cnt[1] += ((PIND & _BV(IN_INJ2_PIN)) >> IN_INJ2_PIN); // injector 2
	}
	for (uint8_t i=0; i<N_CYLS_ACTIVE; i++){ // N of injectors
		if (pin_status_cnt[i] <= INJ_PIN_FILT_LOW){
			INJ_status_filtered[i] = 0; // Low (Injector ON)
		}
		else if (pin_status_cnt[i] >= INJ_PIN_FILT_HIGH){
			INJ_status_filtered[i] = 1; // High (Injector OFF)
		}
	}

	// Manage each injector status
	for (uint8_t i=0; i<N_CYLS_ACTIVE; i++){ // N of injectors = 2
	
		uint8_t output_pin_tmp=OUT_INJ1_PIN; // in case of injector #1
		if (i==1) output_pin_tmp=OUT_INJ2_PIN; // in case of injector #2
	
		// Original ECU is enabling injector (Injector valve will open and gasoline flows)
		if (INJ_status_filtered[i] == 0){ // 0 means that the voltage is GND
			if (INJ_signal_trigger_mode[i] == WAIT_FOR_ON){ // I was waiting for this OFF -> ON transition
				digitalWrite(output_pin_tmp, HIGH); // activates injector
				delta_time_tick[i] = (uint16_t)INJ_exec_time_start - (uint16_t)inj_start_tick_last[i]; // calculates Timer0 ticks (4us) between 2 consecutive injections (2 rpm)
				perc_inc[i] = INJducatimgr.interpolate_rpm_map(delta_time_tick[i]);
				inj_start_tick_last[i] = (uint16_t)INJ_exec_time_start; // saves old time stamp (start of last injection)
				injection_counter[i]++; // increases the combustion cycles
				INJ_signal_trigger_mode[i] = WAIT_FOR_OFF; // Next cycle, wait for ON -> OFF transition
				uint8_t buffer_now_fill_tmp = buffer_last_fill+1; // next buffer fill (temp val)
				if (i==0){ // Only for injector #1
					if (buffer_now_fill_tmp==INJ_COMBUSTION_BUFFERS_NUM) buffer_now_fill_tmp=0; // rollout
					if ((buffer_now_fill_tmp == buffer_now_write) && (buffer_last_write != buffer_now_write)){ // the buffer I am trying to write is being written
						buffer_write_possible = false; // not possible to write
					}
					else{
						buffer_now_fill = buffer_now_fill_tmp; // writing is possible
						combustion_buffers_1[buffer_now_fill].injection_counter = injection_counter[i]; // update buffer
						combustion_buffers_1[buffer_now_fill].inj_start_ticks = INJ_exec_time_start; // update buffer
						buffer_write_possible = true;
					}
				}
			}
		}

		// Original ECU is disabling injector (Injector valve will close and gasoline stops, after Timer1 expires)
		else if (INJ_status_filtered[i] == 1){ // OFF
			if (INJ_signal_trigger_mode[i] == WAIT_FOR_OFF){ // I was waiting for this ON -> OFF transition
				if (buffer_write_possible && (i==0)) combustion_buffers_1[buffer_now_fill].inj_end_in_ticks = INJ_exec_time_start; // update buffer
				bool extension_timer_activated = false; // Timer1 not yet initialized
				uint16_t delta_inj_tick_tmp = (uint16_t)INJ_exec_time_start - (uint16_t)inj_start_tick_last[i]; // delta injetion time, in Timer0 ticks (4 us)
				if ((delta_inj_tick_tmp >= INJ_TIME_TICKS_MIN) && (delta_inj_tick_tmp <= INJ_TIME_TICKS_MAX)) { // injection range check
					if (delta_inj_tick_tmp > INJ_OPENING_TIME_TICKS) { // remove the opening time from the time measured
						delta_inj_tick_tmp -= INJ_OPENING_TIME_TICKS; // real injection time (removed the opening time)
					}
					else{
						delta_inj_tick_tmp = 0; // if time is lower than estimated injector opening time, real injection time is considered 0
					}
					delta_inj_tick[i]=delta_inj_tick_tmp;
					//uint16_t perc_inc_tmp = INJducatimgr.interpolate_tim_map(); // percentage increment depending on injection time (1e16 = 100%)
					//if (perc_inc_tmp > MAX_PERCENTAGE_INJ) perc_inc_tmp = MAX_PERCENTAGE_INJ; // saturates the maximum injection percentage, for safety
					//uint16_t extension_time_ticks_tmp = (uint16_t)(((uint32_t)delta_inj_tick_tmp * (uint32_t)perc_inc_tmp) >> (16 - 3)); // calculates extension time, in Timer1 ticks (0.5us). Need to shift 16 pos (because 1e16 = 100%)
					uint16_t extension_time_ticks_tmp = INJducatimgr.interpolate_tim_map(delta_inj_tick_tmp); // increment depending on injection time
					if ((extension_time_ticks_tmp >= MIN_EXTENSION_TIME_TICKS) && (extension_time_ticks_tmp <= MAX_EXTENSION_TIME_TICKS)){ // Checks if the extension time ticks (Timer1) is acceptable
						if (extension_time_ticks_tmp > INJ_CALCULATION_TIME_TICKS) { // Removes the calculation time requested
							extension_time_ticks_tmp -= INJ_CALCULATION_TIME_TICKS; // real extension time
							Timer1.settimer(extension_time_ticks_tmp, i);
							extension_timer_activated = true; // Timer has been activated
							extension_time_ticks[i] = extension_time_ticks_tmp;
							INJ_signal_trigger_mode[i] = WAIT_FOR_ON; // Waiting for next ON event
						}
					}
				}
				// In case the previous calculaton stopped somewhere before activating the timer, do the following (deactivate injector, and so on)
				if (extension_timer_activated == false){ // timer was not activated, so need to disable the output now
					if (i==0) interrupt_inj_1_off(); // Injector 1
					if (i==1) interrupt_inj_2_off(); // Injector 2
					extension_time_ticks[i] = 0; // no extension time programmed
				}
			}
		}
	
	}
	
	// For execution time measurement
	uint32_t INJ_exec_time_end = Timer0_tick_counts(); // 1 tick = 4us
	INJ_exec_time = (uint8_t)INJ_exec_time_end - (uint8_t)INJ_exec_time_start; // Measure interrupt execution time

}

INJducatimgr_class INJducatimgr;

#endif