#ifndef SDmgr_cpp
#define SDmgr_cpp

#include <SDFatYield.h> // SD FAT management (modified SDFat library)
#include "../EEPROMmgr/EEPROMmgr.h" // Need to manage reading from EEPROM (file name, ...)
#include "../INJducatimgr/INJducatimgr.h" // INJ manager. To have access to Engine variables (inj time, ...)
#include "../COMMmgr/COMMmgr.h" // for the checksum calculation
#include "../ADCmgr/ADCmgr.h" // for Battery Reading
#include "../compile_options.h" // Compile options
#include "SDmgr.h"

#define SD_CS_PIN_NUM A1 // CS pin for SD card (SPI) - Slave Select
#define MAX_WRITE_ERRORS 100 // Maximum consecutive write errors that will cause a re-init. 
#define MAX_DELAY_POWERON 50 // This is used as time delay at power ON before performing an "init"
#define COM_MAX_BUFFER_WRITE_PER_CALL 5 // max

SdFat SD; // needed to manage SD card

// Date for SD logging
uint16_t GPS_year = 1987; // dummy
uint8_t GPS_month = 11; // dummy
uint8_t GPS_day = 6; // dummy
uint8_t GPS_hour = 8; // dummy
uint8_t GPS_min = 0; // dummy
uint8_t GPS_sec = 0; // dummy

// call back for file timestamps
void dateTime(uint16_t* date, uint16_t* time) {

  // return date using FAT_DATE macro to format fields
  //*date = FAT_DATE(now.year(), now.month(), now.day());
  *date = FAT_DATE(GPS_year, GPS_month, GPS_day);

  // return time using FAT_TIME macro to format fields
  //*time = FAT_TIME(now.hour(), now.minute(), now.second());
  *time = FAT_TIME(GPS_hour, GPS_min, GPS_sec);
}


SDmgr_class::SDmgr_class(){
	
	SD_init_OK = false; // NG
	write_errors_cnt = 0;
	
}


bool SDmgr_class::begin(){
	
#if SD_MODULE_PRESENT
	
	if (!ADCmgr_battery_status_read() && !bat_check_inhibit()) return false; // Return "False" in case battery was OFF, avoiding SD initialization
	
	// Set date time callback function
	SdFile::dateTimeCallback(dateTime);
	
	// SD INITIALIZATION
	if (!SD.begin(SD_CS_PIN_NUM, SPI_HALF_SPEED)) { // CS pin for SD card is pin #10
		SD_init_OK = false; // NG
		return false;
	}
	SD_init_OK = true; // OK

	// Read file name from EEPROM and stores file name as string (8.3 format)
	uint16_t file_number = EEPROM_SD_file_num_rw();
	file_name="fln";
	if (file_number < 10) file_name+='0';
	if (file_number < 100) file_name+='0';
	if (file_number < 1000) file_name+='0';
	if (file_number < 10000) file_name+='0';
	file_name+=(unsigned int)file_number;
	file_name+=".log";
	
	//MPU6050mgr.flush_buffer(); // flushes the IMU buffer (sets no data to write)
	
#endif
	
	return true; // Init successful
	
}


bool SDmgr_class::log_SD_data(){

	bool temp_reply = false; // temporary response

#if SD_MODULE_PRESENT

	if (all_log_inhibit()) return temp_reply; // return "False" in case of Log Inhibit (EEPROM), avoiding SD initialization and writing

	// Performs SD Initialization, in case it is necessary (in case it was not possible before, or it is the first time to call this function)
	if (SD_init_OK == false) {
		write_errors_cnt++; // this is for delay purposes, to avoid continuous "begin()" calls
		if (write_errors_cnt >= MAX_DELAY_POWERON) {
			begin(); // Initializes SD memory
			write_errors_cnt = 0; // reset counter
		} 
	}

	// Writing the data on SD memory
	if (SD_init_OK == true){ // SD card initialized properly
	
		if (ADCmgr_battery_status_read() || bat_check_inhibit()){ // logs only if Battery is ON (or if battery check inhibit config flag is active)
			File dataFile = SD.open(file_name, FILE_WRITE); // Opens the file only if there is battery voltage
			if (dataFile){ // log only when the file is open, and when battery is connected (to avoid memory corruption)
				
				uint8_t data_bytes_written; // Data bytes effectively written on the SD card
				bool error_status = false; // Error status

				// Engine info
				/*if (!eng_log_inhibit()){
					data_bytes_written = dataFile.write(INJducatimgr.SD_writing_buffer, INJ_ENG_DATA_SD_COMM_BUFFER_SIZE); // Writes engine related info (calculated above)
					if (data_bytes_written != INJ_ENG_DATA_SD_COMM_BUFFER_SIZE) error_status = true; // not all bytes were written
				}*/

				// COMBUSTION data
				uint8_t com_buffer_write_cnt = 0; // initialize to 0
				while ((buffer_last_write != buffer_last_fill) && (com_buffer_write_cnt < COM_MAX_BUFFER_WRITE_PER_CALL)){ // there is a new filled in data
					INJducatimgr.create_combustion_data_packet(); // creates the combustion data packet (SD_combustion_buffer)
					data_bytes_written = dataFile.write(INJducatimgr.SD_combustion_buffer, INJ_COMB_DATA_SD_COMM_BUFFER_SIZE); // Writes combustion related info (calculated above)
					buffer_last_write = buffer_now_write; // writing has been done completely
					if (data_bytes_written != INJ_COMB_DATA_SD_COMM_BUFFER_SIZE) error_status = true; // not all bytes were written
					com_buffer_write_cnt++; // increase counter
				}

				// Error status check
				if (error_status == true) { // Could not write completely all data
					write_errors_cnt++; // Increase error counter
				}else{ // Completely written all data
					write_errors_cnt = 0; // No error
				}
				
				if (ADCmgr_battery_status_read() || bat_check_inhibit()) { // Data written only in case battery voltage is present, or if bypass flag is active
					dataFile.close(); // Really write the data on the SD, only if battery voltage is present
					if (error_status == false) { // No error found
						temp_reply = true; // Data log considered completed successfully
					}
				}
				
			}else{ // Could not open the file
				write_errors_cnt++; // Increase error counter
			}
		}
		
		// Errors max check
		if (write_errors_cnt >= MAX_WRITE_ERRORS) {
			SD_init_OK = false; // This will cause the SD card to be re-initialized at next function call
			write_errors_cnt = 0; // Reset error counter
		}
		
	}
#endif
	
	return temp_reply;
	
}

SDmgr_class SDmgr;

#endif